/*---------------------------------------------------------------------
  encoder_demo by Jack Botner
  
  To provide a demonatration program for the Greyhill 62P22-L4
  pushbutton rotary encoder.

  Processor: PIC16F84A
  Clock: 3.58 mHz crystal
---------------------------------------------------------------------*/
#include <system.h>
#include "encoder_demo.h"

#pragma CLOCK_FREQ      3579545
#pragma DATA _CONFIG, _CP_OFF & _HS_OSC & _PWRTE_ON & _WDT_OFF

// Global Variables

static volatile unsigned char ucEventFlags=0;		// event flags
static volatile unsigned char ucDebouncedState=0;	// debounced state of the switches
static volatile unsigned char ucPreviousState=0;	// previous switch state

static volatile char led_number=0;			// number on the LED
static volatile unsigned char beep_delay=0; // beep interval, 4ms ticks

// Internal functions

static void analyze_switch(void);
static void debounce_switch(void);
static void InitCPU(void);
static void ledOutputNumber( char cNumber );

/*----------------------------------------------------------------
  main()
----------------------------------------------------------------*/
void main()
{
  InitCPU();			// initialize hardware and start timer0

  ledOutputNumber( led_number );	// display '0'

  while (1)
  {
	if ( test_bit( ucEventFlags, EVENT_SWITCH_CHANGE ) )
    {
      // After debouncing a switch change was detected.
      // Find out what it was and generate the corresponding event.
      analyze_switch();       // convert switch change to events
      clear_bit( ucEventFlags, EVENT_SWITCH_CHANGE );
    }
      
    if ( test_bit( ucEventFlags, EVENT_ENCODER_UP ) )
    {
      // The encoder was turned clockwise one step.
	  if ( led_number == 9 )
	    led_number = 0;
	  else
	    ++led_number;  
	  ledOutputNumber( led_number );
      clear_bit( ucEventFlags, EVENT_ENCODER_UP );
    }

    if ( test_bit( ucEventFlags, EVENT_ENCODER_DOWN ) )
    {
      // The encoder was turned counterclockwise one step.
	  if ( led_number )
	    --led_number;
	  else
	    led_number = 9;
	  ledOutputNumber( led_number );
      clear_bit( ucEventFlags, EVENT_ENCODER_DOWN );
    }  

    if ( test_bit( ucEventFlags, EVENT_BUTTON_PRESSED ) )
    { 
	  // The encoder button was pressed.
	  set_bit( BEEPER_PORT, BEEPER_BIT ); 	// sound beeper
	  beep_delay = 75;		// schedule it to turn off (75x4ms)
      clear_bit( ucEventFlags, EVENT_BUTTON_PRESSED );
    }
  }    
}

/*---------------------------------------------------------------------
  interrupt()
  
  Control comes here for a system interrupt. The only interrupt
  implemented is timer0, every 4 ms.
---------------------------------------------------------------------*/
void interrupt( void )
{
  if ( test_bit( intcon, T0IF ) )	// timer0 interrupt
  {
	// Timer0 is used to generate 4 ms interrupts for switch debouncing.
    // Poll the switches and update the debounce state data. 
    debounce_switch();			// sets ucDebouncedState
    if ( ucDebouncedState != ucPreviousState )
	  set_bit( ucEventFlags, EVENT_SWITCH_CHANGE );
    
    // Manage the length of time the beep sounds.
    if ( beep_delay )
    {
      --beep_delay;
      if ( !beep_delay )
        // This should be an event for the mainline but it is too simple.
        clear_bit( BEEPER_PORT, BEEPER_BIT );
    }

    // Reset the timer
    tmr0 = TIMER0_INIT;			// reload the counter
    clear_bit( intcon, T0IF );	// reset the TMR0 overflow interrupt flag
  }
}

/*-----------------------------------------------------------------
  debounce_switch()

  Called every 4 ms from the mainline, debounces inputs from
  the rotary encoder.
  
  The technique documented in "A Guide to Debouncing" by Jack
  Ganssle is used to debounce up to 8 inputs simultaneously
  (3 in this case).

  Output: ucDebouncedState
-----------------------------------------------------------------*/
static void debounce_switch()
{
  static volatile char ucFirst=0;        		// encoder state managment
  static volatile unsigned char ucIndex=0;              // index into ucState
  static volatile unsigned char ucState[MAX_CHECKS] =   // array containing bounce status
                { 0, 0, 0, 0, 0 };

  unsigned char ii, ucResult, ucSwIn=0;

  // Read current switch states
  if ( !test_bit( ENCODER_PORT, ENCODER_A ) )
    set_bit( ucSwIn, ENC_SWITCH_A );
  if ( !test_bit( ENCODER_PORT, ENCODER_B ) )
    set_bit( ucSwIn, ENC_SWITCH_B );
  if ( !test_bit( ENCODER_PORT, ENCODER_PB ) )
    set_bit( ucSwIn, ENC_SWITCH_PB );

  ucState[ucIndex] = ucSwIn;

  ++ucIndex;
  if ( ucIndex == MAX_CHECKS )
    ucIndex = 0;

  // And all columns of the array
  for ( ii=0, ucResult=0xff; ii<(MAX_CHECKS-1); ++ii )
    ucResult = ucResult & ucState[ii];

  ucDebouncedState = ucResult;

  // Start of execution only: Until the debounce routine settles,
  // set the previous switch state to the current switch state.
  if ( ucFirst < MAX_CHECKS )
  {
    ucPreviousState = ucDebouncedState;
    ++ucFirst;
  }
}

/*-----------------------------------------------------------------
  analyze_switch()

  Called when the state of the switch changes.
  Analyzes the rotary encoder to determine if the shaft rotation
  is up or down, or if the pushbutton has been pressed, and
  generate the corresponding event.

  Output: various ucEventFlags
-----------------------------------------------------------------*/
static void analyze_switch()
{
  unsigned char   ucNow, ucThen;

  if ( test_bit( ucDebouncedState, ENC_SWITCH_PB ) &&
	!test_bit( ucPreviousState, ENC_SWITCH_PB ) )
  {
    // The encoder button has been pressed
	set_bit( ucEventFlags, EVENT_BUTTON_PRESSED );
  }

  // Convert the position of the encoder to a number 0..3
  ucNow = ucDebouncedState & 0x03;
  ucThen = ucPreviousState & 0x03;

  // Given the A output in bit 0 and B output in bit 1 positions, the
  // encoder generates the values 0, 1, 3, 2 as it is rotated clockwise.

  if ( ucNow == 0 )
  {
    if ( ucThen == 2 )
	  set_bit( ucEventFlags, EVENT_ENCODER_UP );
	else
	  if ( ucThen == 1 )
		set_bit( ucEventFlags, EVENT_ENCODER_DOWN );
  }
  else
  {
	if ( ucNow == 1 )
	{
      if ( ucThen == 0 )
		set_bit( ucEventFlags, EVENT_ENCODER_UP );
	  else
		if ( ucThen == 3 )
		  set_bit( ucEventFlags, EVENT_ENCODER_DOWN );
	}
	else
	{
	  if ( ucNow == 3 )
	  {
        if ( ucThen == 1 )
		  set_bit( ucEventFlags, EVENT_ENCODER_UP );
		else
		  if ( ucThen == 2 )
			set_bit( ucEventFlags, EVENT_ENCODER_DOWN );
	  }
	  else  // ucNow == 2
	  {
        if ( ucThen == 3 )
			set_bit( ucEventFlags, EVENT_ENCODER_UP );
		else
		  if ( ucThen == 0 )
			set_bit( ucEventFlags, EVENT_ENCODER_DOWN );
	  }
	}
  }	  

  ucPreviousState = ucDebouncedState;	// reset state
}

/*---------------------------------------------------------------------
  ledOutputNumber(...)

  Outputs a number to the led display. Clearing the port turns
  the corresponding segment on.

  Input: a single numeric digit, range 0..9 (ascii or binary, only the
         low-order nibble is processed.) Also the value of x'0f'
		 which means blank the display.
---------------------------------------------------------------------*/

// Bit positions in the table of segments
#define SEG_A   0
#define SEG_B   1
#define SEG_C   2
#define SEG_D   3
#define SEG_E   4
#define SEG_F   5
#define SEG_G   6

static void ledOutputNumber( char cNumber )
{
  // Segment A = bit 0, and so on, for each digit 0..9
  static const unsigned char ucDigTab[10] =
  {
	0b00111111, 0b00000110, 0b01011011, 0b01001111, 0b01100110,
	0b01101101, 0b01111101, 0b00000111, 0b11111111, 0b01100111
  };

  unsigned char	ucNibble, ucSegments;

  ucNibble = (unsigned char) cNumber & 0x0f;
  ucSegments = ( ucNibble > 9 ) ? 0 : ucDigTab[ucNibble];

  if ( test_bit( ucSegments, SEG_A ) )
    clear_bit( SEG_A_PORT, SEG_A_BIT );
  else
    set_bit( SEG_A_PORT, SEG_A_BIT );

  if ( test_bit( ucSegments, SEG_B ) )
    clear_bit( SEG_B_PORT, SEG_B_BIT );
  else
    set_bit( SEG_B_PORT, SEG_B_BIT );

  if ( test_bit( ucSegments, SEG_C ) )
    clear_bit( SEG_C_PORT, SEG_C_BIT );
  else
    set_bit( SEG_C_PORT, SEG_C_BIT );

  if ( test_bit( ucSegments, SEG_D ) )
    clear_bit( SEG_D_PORT, SEG_D_BIT );
  else
    set_bit( SEG_D_PORT, SEG_D_BIT );

  if ( test_bit( ucSegments, SEG_E ) )
    clear_bit( SEG_E_PORT, SEG_E_BIT );
  else
    set_bit( SEG_E_PORT, SEG_E_BIT );

  if ( test_bit( ucSegments, SEG_F ) )
    clear_bit( SEG_F_PORT, SEG_F_BIT );
  else
    set_bit( SEG_F_PORT, SEG_F_BIT );

  if ( test_bit( ucSegments, SEG_G ) )
    clear_bit( SEG_G_PORT, SEG_G_BIT );
  else
    set_bit( SEG_G_PORT, SEG_G_BIT );
}

/*---------------------------------------------------------------------
  InitCPU()
---------------------------------------------------------------------*/
static void InitCPU()
{
  // Initialize PORTA
  // PORTA connections as follows:
  //  RA0 - (output) DSP1 'a' segment
  //  RA1 - (output) DSP1 'b' segment
  //  RA2 - (output) DSP1 'c' segment
  //  RA3 - (output) DSP1 'd' segment
  //  RA4 - (output) DSP1 'e' segment

  trisa = 00000000b;            // 1=input, 0=output
  porta = 00011111b;			// led segments all off

  // Initialize PORTB
  // PORTB connections as follows:
  //  RB0 - (input) Rotary Encoder pushbutton *
  //  RB1 - (input) Rotary Encoder OUT A *
  //  RB2 - (input) Rotary Encoder OUT B *
  //  RB3 - (output) DSP1 'f' segment
  //  RB4 - (output) DSP1 'g' segment
  //  RB5 - (output) piezo buzzer (or LED)
  //  RB6 - (input) (ICSPCLK)
  //  RB7 - (input) (ICSPDAT)
  // * = pull-up required

  trisb = 11000111b;            	// 1=input, 0=output
  portb = 00011000b;				// led segments all off
  
  // Initialize timer0 to produce an interrupt every 4 ms.
  // Timer0 counts up 8 bits, giving a maximum of 256 counts. Counting
  // is done by the processor, each instruction cycle, which is 4 clock
  // cycles. Since the clock is 3.58 mHz, the timer increments every 
  // 1.11 microsecond. With 16:1 prescaler we need to count 225 to get
  // 4 ms. This means initializing the timer with 256-225=31.
  
								// weak pull-ups on port B inputs
  option_reg = 00000011b;		// prescaler=16
  tmr0 = TIMER0_INIT;

  // Enable interrupts for Timer0

  set_bit( intcon, GIE );       // enable global interrupts
  set_bit( intcon, T0IE );      // enable timer0 interrupts
}
